{**********************************************}
{    Borland Delphi Unit                       }
{                                              }
{    TMatrix and TMatrixViewer implementation  }
{                                              }
{    Created 1996 by Vladimir Safin            }
{    e-mail:  vlads@uniyar.ac.ru               }
{                                              }
{**********************************************}

unit Matrix;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, Grids, StdCtrls;

type
  { Array of Extended }
  PExtendedArray = ^TExtendedArray;
  TExtendedArray = array [0..255] of Extended;
  { Array of pointers }
  TPtrArray = array [0..255] of PExtendedArray;

const
  { Size of Extended }
  szExtended = SizeOf(Extended);

type
  EMatrixException = class(Exception);

  TMatrix = class;

  { Viewer and editor for TMatrix }
  TMatrixViewer = class(TCustomGrid)
  private
    FAdjustSize: Boolean;
    FDigits: Integer;
    FMatrix: TMatrix;
    procedure Resize;
    procedure SetAdjustSize(Value: Boolean);
    procedure SetDigits(Value: Integer);
  protected
    { Protected declarations }
    function  CanEditAcceptKey(Key: Char): Boolean; override;
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect;
                       AState: TGridDrawState);  override;
    function  GetEditText(ACol, ARow: Longint): string; override;
    procedure Paint;  override;
    procedure SetEditText(ACol, ARow: Longint; const Value: string); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy;       override;
    procedure   SetMatrix(Value: TMatrix);
    procedure   RefreshMatrix;
  published
    { Published declarations }
    property AdjustSize: Boolean read FAdjustSize write SetAdjustSize;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DefaultColWidth;
    property DefaultRowHeight;
    property Digits  : Integer read FDigits  write SetDigits default 2;
    property Font;
    property GridLineWidth;
    property Options;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;


  { TMatrix - the table Lines x Columns  }
  TMatrix = class(TComponent)
  private
    FRows: TPtrArray;
    FColumns: Longint;
    FLines: Longint;
    FViewer: TMatrixViewer;
    function  GetCells(ARow, ACol: Integer): Extended;
    procedure SetCells(ARow, ACol: Integer; Value: Extended);
    procedure SetColumns(Value: Longint);
    procedure SetLines  (Value: Longint);
    procedure SetViewer (Value: TMatrixViewer);
  public
    constructor Create(AOwner: TComponent);   override;
    destructor  Destroy;                      override;
    procedure   Assign(Source: TPersistent);  override;
    procedure   AssignValue(Value: Extended);
    procedure   AssignToLine(NLine: Byte; Values: Array of Extended);
    procedure   AssignToCol (NCol : Byte; Values: Array of Extended);
    procedure   AddMatrix(AMatrix: TMatrix);
    procedure   MultMatrix(AMatrix: TMatrix);
    procedure   MultValue(Value: Extended);
    procedure   DiagMatrix(Value: Extended);
    function    Determinant: Extended;
    procedure   Inverse;
    procedure   Transpose;
    procedure   Repaint;
    procedure   ChangeLines(i, j: Integer);
    procedure   AddLines(i, j: integer; Factor: Extended);
    procedure   MultLine(i: Integer; Factor: Extended);
    property    Cells[ARow, ACol: Integer]: Extended read GetCells write SetCells;
  published
    property Columns: Longint read FColumns write SetColumns;
    property Lines: Longint read FLines write SetLines;
    property Viewer: TMatrixViewer read FViewer write SetViewer;
  end;


function M_Inverse(AMatrix: TMatrix): TMatrix;
function M_SolveLinearSystem(A, B: TMatrix): TMatrix;
function M_MultMatrix(A, B: TMatrix): TMatrix;
function M_AddMatrix(A, B: TMatrix): TMatrix;
function M_MultValue(A: TMatrix; Value: Extended): TMatrix;


procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Math', [TMatrix, TMatrixViewer]);
end;

function SolveMatrix(AMatrix: TMatrix; var BMatrix: TMatrix): Extended;
var
  i, j: Integer;
  Temp: Extended;
  Factor: Integer;
begin
  Result := 1;
  for i := 0 to AMatrix.Lines - 1 do
  begin
    j := 0;
    while j < AMatrix.Lines do
      if AMatrix.Cells[i, j] <> 0 then break
                                  else inc(j);
    if j = AMatrix.Lines then
    begin
      Result := 0;
      Exit;
    end;
    AMatrix.ChangeLines(i, j);
    BMatrix.ChangeLines(i, j);
    if not Odd(AMatrix.Columns) then Result := -1 * Result;
    for j := i + 1 to AMatrix.Lines - 1 do
    begin
      Temp := AMatrix.Cells[j, i]/AMatrix.Cells[i, i];
      AMatrix.AddLines(j, i, -Temp);
      BMatrix.AddLines(j, i, -Temp);
    end;
  end;

  for i := AMatrix.Lines - 1 downto 0 do
    for j := i - 1 downto 0 do
    begin
      Temp := AMatrix.Cells[j, i]/AMatrix.Cells[i, i];
      AMatrix.AddLines(j, i, -Temp);
      BMatrix.AddLines(j, i, -Temp);
    end;

  for i := 0 to AMatrix.Lines - 1 do
  begin
    Result := Result * AMatrix.Cells[i, i];
    Temp := 1/AMatrix.Cells[i, i];
    BMatrix.MultLine(i, Temp);
    AMatrix.MultLine(i, Temp);
  end;
end;

function M_Inverse(AMatrix: TMatrix): TMatrix;
var
  A: TMatrix;
begin
  if AMatrix.FLines <> AMatrix.FColumns
    then raise EMatrixException.Create('This operation can be applied only to' +
                                       ' square matrix');
  A := TMatrix.Create(nil);
  A.Assign(AMatrix);

  Result := TMatrix.Create(nil);
  Result.Lines := AMatrix.Lines;
  Result.Columns := AMatrix.Columns;
  Result.DiagMatrix(1);

  if SolveMatrix(A, Result) = 0 then
  begin
    raise EMatrixException.Create('Could not inverse this matrix');
    Result := nil;
  end;
end;

function M_SolveLinearSystem(A, B: TMatrix): TMatrix;
var
  C: TMatrix;
begin
  if (A.FLines <> A.FColumns) or (B.Lines <> A.Lines) or (B.Columns <> 1)
    then raise EMatrixException.Create('It is not a linear system');

  C := TMatrix.Create(nil);
  C.Assign(A);

  Result := TMatrix.Create(nil);
  Result.Assign(B);

  if SolveMatrix(C, Result) = 0 then
  begin
    raise EMatrixException.Create('Could not solve this system');
    Result := nil;
  end;

end;

function M_MultMatrix(A, B: TMatrix): TMatrix;
begin
  Result := TMatrix.Create(nil);
  Result.Assign(A);
  Result.MultMatrix(B);
end;

function M_AddMatrix(A, B: TMatrix): TMatrix;
begin
  Result := TMatrix.Create(nil);
  Result.Assign(A);
  Result.AddMatrix(B);
end;

function M_MultValue(A: TMatrix; Value: Extended): TMatrix;
begin
  Result := TMatrix.Create(nil);
  Result.Assign(A);
  Result.MultValue(Value);
end;

{ -- TMatrix -- }
constructor TMatrix.Create;
var
  i: integer;
begin
  inherited Create(AOwner);
  { default 3x3 }
  { elements are created automaticaly when you change Lines and Columns }
  Lines := 3;
  Columns := 3;
end;

destructor TMatrix.Destroy;
var
  I: Integer;
begin
  if Viewer <> nil then Viewer.SetMatrix(nil);
  { deallocate memory }
  for i := 0 to FLines - 1 do
    FreeMem(FRows[i], szExtended * FColumns);
  inherited Destroy;
end;

procedure TMatrix.Assign(Source: TPersistent);
var
  i, j: Byte;
begin
  Lines := (Source as TMatrix).Lines;
  Columns := (Source as TMatrix).Columns;
  for i := 0 to FLines - 1 do
    for j := 0 to FColumns - 1 do
      FRows[i]^[j] := (Source as TMatrix).Cells[i, j];
  Repaint;
end;

procedure TMatrix.AssignValue(Value: Extended);
var
  i, j: Integer;
begin
  for i := 0 to FLines - 1 do
    for j := 0 to FColumns - 1 do
      FRows[i]^[j] := Value;
  Repaint;
end;

procedure TMatrix.AssignToLine(NLine: Byte; Values: Array of Extended);
var
  j, Min: Integer;
begin
  if NLine >= FLines then exit;
  if FColumns > High(Values)
    then Min := High(Values)
    else Min := FColumns - 1;
  for j := 0 to Min do
    FRows[NLine]^[j] := Values[j];
  Repaint;
end;

procedure TMatrix.AssignToCol(NCol: Byte; Values: Array of Extended);
var
  j, Min: Integer;
begin
  if NCol >= FColumns then exit;
  if FLines > High(Values)
    then Min := High(Values)
    else Min := FLines - 1;
  for j := 0 to Min do
    FRows[j]^[NCol] := Values[j];
  Repaint;
end;

procedure TMatrix.AddMatrix(AMatrix: TMatrix);
var
  i, j: Integer;
begin
  if (Columns <> AMatrix.Columns) or (Lines <> AMatrix.Lines)
    then raise EMatrixException.Create('Can not add these matrixes');
  for I := 0 to FLines - 1 do
    for j := 0 to FColumns - 1 do
      FRows[i]^[j] := FRows[i]^[j] + AMatrix.Cells[i, j];
  Repaint;
end;

procedure TMatrix.MultMatrix(AMatrix: TMatrix);
var
  A: TMatrix;
  i, j, k: Integer;
begin
  if Columns <> AMatrix.Lines
    then raise EMatrixException.Create('Can not multiply these matrixes');
  A := TMatrix.Create(nil);
  A.Lines := Lines;
  A.Columns := AMatrix.Columns;
  for k := 0 to Lines - 1 do
    for i := 0 to AMatrix.Columns - 1 do
      for j := 0 to Columns - 1 do
        A.Cells[k, i] := A.Cells[k, i] + Cells[k, j] * AMatrix.Cells[j, i];
  Assign(A);
  A.Free;
  Repaint;
end;

procedure TMatrix.MultValue(Value: Extended);
var
  i, j: Byte;
begin
  for i := 0 to FLines - 1 do
    for j := 0 to FColumns - 1 do
      FRows[i]^[j] := FRows[i]^[j] * Value;
  Repaint;
end;

procedure TMatrix.DiagMatrix(Value: Extended);
var
  i: Integer;
begin
  if FLines <> FColumns
    then raise EMatrixException.Create('This operation can be applied only to' +
                                       ' square matrix');
  AssignValue(0);
  for i := 0 to FLines - 1 do
    FRows[i]^[i] := Value;
  Repaint;
end;

function TMatrix.Determinant: Extended;
var
  A, B: TMatrix;
begin
  if FLines <> FColumns
    then raise EMatrixException.Create('This operation can be applied only to' +
                                       ' square matrix');
  A := TMatrix.Create(nil);
  A.Assign(Self);

  B := TMatrix.Create(nil);
  B.Lines := Lines;
  B.Columns := Columns;
  B.DiagMatrix(1);

  Result := SolveMatrix(A, B);
end;

procedure TMatrix.Inverse;
var
  A: TMatrix;
begin
  if FLines <> FColumns
    then raise EMatrixException.Create('This operation can be applied only to' +
                                       ' square matrix');
  A := TMatrix.Create(nil);
  A.Lines := Lines;
  A.Columns := Columns;
  A.DiagMatrix(1);

  if SolveMatrix(Self, A) = 0 then
  begin
    raise EMatrixException.Create('Could not inverse this matrix');
    Exit;
  end;
  Assign(A);
  A.Repaint;
end;

procedure TMatrix.Transpose;
var
  T: Extended;
  i, j: Integer;
begin
  if FLines <> FColumns
    then raise EMatrixException.Create('This operation can be applied only to' +
                                       ' square matrix');
  for i := 0 to FLines - 1 do
    for j := i + 1 to FColumns - 1 do
    begin
      T := FRows[i]^[j];
      FRows[i]^[j] := FRows[j]^[i];
      FRows[j]^[i] := T;
    end;
  Repaint;
end;

function  TMatrix.GetCells(ARow, ACol: Integer): Extended;
begin
  { if index is invalid then raise exception }
  if (ACol < 0) or (ACol > FColumns - 1) or
     (ARow < 0) or (ARow > FLines   - 1)
  then  raise EMatrixException.Create('Index out of bounds');
  Result := FRows[ARow]^[ACol];
end;

procedure TMatrix.SetCells(ARow, ACol: Integer; Value: Extended);
begin
  { if index is invalid then raise exception }
  if (ACol < 0) or (ACol > FColumns - 1) or
     (ARow < 0) or (ARow > FLines - 1)
  then  raise EMatrixException.Create('Index out of bounds');
  FRows[ARow]^[ACol] := Value;
  Repaint;
end;

procedure TMatrix.SetColumns(Value: Longint);
var
  OldColumns: Longint;
  i: Integer;
begin
  if FColumns <> Value then begin
    { columns >= 1 }
    if Value < 1 then Value := 1;
    OldColumns := FColumns;
    { reallocate memory }
    { if OldColumns < Value then new elements will be equal to 0 }
    { if OldColumns > Value then elements will be truncate }
    for i := 0 to FLines - 1 do
{$IFDEF WIN32}
      ReAllocMem(FRows[i], szExtended * Value);
{$ELSE}
      FRows[i] := ReAllocMem(FRows[i], szExtended * OldColumns,
                                       szExtended * Value);
{$ENDIF}
    { update FColumns }
    FColumns := Value;
    Repaint;
  end
end;

procedure TMatrix.SetLines(Value: Longint);
var
  i: Integer;
begin
  if Value = FLines then Exit;
  { Lines >= 1 }
  if Value < 1 then Value := 1;
  { if Value > Lines, new linesw will be added }
  if Value > FLines then
    for i := FLines to Value - 1 do
      FRows[i] := AllocMem(szExtended * FColumns)
  { if Value < Lines, unnessesary lines will be destroed }
  else if Value < FLines then
    for i := FLines - 1 downto Value do
      FreeMem(FRows[i], szExtended * FColumns);
  { update FLines }
  FLines := Value;
  Repaint;
end;

procedure TMatrix.SetViewer (Value: TMatrixViewer);
begin
  FViewer := Value;
  if Value <> nil then
  begin
    Value.SetMatrix(Self);
    Repaint;;
  end;
end;

procedure TMatrix.Repaint;
begin
  if FViewer <> nil then FViewer.RefreshMatrix;
end;

procedure TMatrix.ChangeLines(i, j: Integer);
var
  k: Integer;
  Temp: Extended;
begin
  for k := 0 to FColumns - 1 do
  begin
    Temp := FRows[i]^[k];
    FRows[i]^[k] := FRows[j]^[k];
    FRows[j]^[k] := Temp;
  end;
  Repaint;
end;

procedure TMatrix.AddLines(i, j: integer; Factor: Extended);
var
  k: Integer;
begin
  for k := 0 to FColumns - 1 do
    FRows[i]^[k] := FRows[i]^[k] + FRows[j]^[k] * Factor;
  Repaint;
end;

procedure TMatrix.MultLine(i: Integer; Factor: Extended);
var
  k: Integer;
begin
  for k := 0 to FColumns - 1 do
    FRows[i]^[k] := FRows[i]^[k] * Factor;
  Repaint;
end;

{ -- TMatrixViewer -- }
constructor TMatrixViewer.Create(AOwner: TComponent);
var
  i: Integer;
begin
  inherited Create(AOwner);
  FixedRows := 0;
  FixedCols := 0;
  ScrollBars := ssNone;
  Options := Options + [goEditing];
  FMatrix := nil;
  Digits := 2;
  AdjustSize := False;
end;

destructor TMatrixViewer.Destroy;
begin
  if FMatrix <> nil then FMatrix.Viewer := nil;
  inherited Destroy;
end;

procedure TMatrixViewer.SetMatrix(Value: TMatrix);
begin
  FMatrix := Value;
  if Value <> nil then
  begin
    ColCount := Value.Columns;
    RowCount := Value.Lines;
    Resize;
    Invalidate;
  end;
end;

function TMatrixViewer.GetEditText(ACol, ARow: Longint): string;
begin
  if FMatrix <> nil
    then Result := FloatToStrF(FMatrix.Cells[ARow, ACol], ffFixed, 10, Digits)
    else Result := '';
end;

procedure TMatrixViewer.SetEditText(ACol, ARow: Longint; const Value: string);
var
  V: Extended;
  Er: Integer;
begin
  if FMatrix = nil then Exit;
  Val(Value, V, Er);
  if Er <> 0 then Exit;
  FMatrix.Cells[ARow, ACol] := V;
  inherited SetEditText(ACol, ARow, Value);
end;

function TMatrixViewer.CanEditAcceptKey(Key: Char): Boolean;
begin
  Result := Key in ['0'..'9', '-', DecimalSeparator];
end;

procedure TMatrixViewer.Paint;
begin
  if FMatrix <> nil then Resize;
  inherited Paint;
end;

procedure TMatrixViewer.DrawCell(ACol, ARow: Longint; ARect: TRect;
  AState: TGridDrawState);
var
  Text: string;
begin
  if FMatrix = nil then Text := ''
  else Text := FloatToStrF(FMatrix.Cells[ARow, ACol], ffFixed, 10, Digits);
  ExtTextOut(Canvas.Handle, ARect.Left + 2, ARect.Top + 2, ETO_CLIPPED or
    ETO_OPAQUE, @ARect, @Text[1], Length(Text), nil);
end;


procedure TMatrixViewer.Resize;
begin
  if FMatrix = nil then exit;
  if FAdjustSize then SetBounds(Left, Top, FMatrix.Columns * (DefaultColWidth + 1) + 1,
                                       FMatrix.Lines * (DefaultRowHeight + 1) + 1);
end;

procedure TMatrixViewer.SetAdjustSize(Value: Boolean);
begin
  FAdjustSize := Value;
  Resize;
end;

procedure TMatrixViewer.SetDigits(Value: Integer);
begin
  if FDigits <> Value then
  begin
    FDigits := Value;
    Invalidate;
  end
end;

procedure TMatrixViewer.RefreshMatrix;
begin
  if FMatrix <> nil then
  begin
    ColCount := FMatrix.Columns;
    RowCount := FMatrix.Lines;
    Invalidate;
  end;
end;

end.
